@@ -930,7 +930,7 @@ GDALDataset *ERSDataset::Open(GDALOpenInfo *poOpenInfo)
     /* -------------------------------------------------------------------- */
     /*      Create a corresponding GDALDataset.                             */
     /* -------------------------------------------------------------------- */
-    auto poDS = cpl::make_unique<ERSDataset>();
+    auto poDS = std::make_unique<ERSDataset>();
     poDS->poHeader = poHeader;
     poDS->eAccess = poOpenInfo->eAccess;
 
@@ -1057,7 +1057,7 @@ GDALDataset *ERSDataset::Open(GDALOpenInfo *poOpenInfo)
         else
             poDS->fpImage = VSIFOpenL(osDataFilePath, "r");
 
-        poDS->osRawFilename = osDataFilePath;
+        poDS->osRawFilename = std::move(osDataFilePath);
 
         if (poDS->fpImage != nullptr && nBands > 0)
         {
@@ -1075,14 +1075,15 @@ GDALDataset *ERSDataset::Open(GDALOpenInfo *poOpenInfo)
             if (!RAWDatasetCheckMemoryUsage(
                     poDS->nRasterXSize, poDS->nRasterYSize, nBands, iWordSize,
                     iWordSize, iWordSize * nBands * poDS->nRasterXSize,
-                    nHeaderOffset, iWordSize * poDS->nRasterXSize,
+                    nHeaderOffset,
+                    static_cast<vsi_l_offset>(iWordSize) * poDS->nRasterXSize,
                     poDS->fpImage))
             {
                 return nullptr;
             }
-            if (nHeaderOffset >
-                std::numeric_limits<GIntBig>::max() -
-                    (nBands - 1) * iWordSize * poDS->nRasterXSize)
+            if (nHeaderOffset > std::numeric_limits<GIntBig>::max() -
+                                    static_cast<GIntBig>(nBands - 1) *
+                                        iWordSize * poDS->nRasterXSize)
             {
                 CPLError(CE_Failure, CPLE_AppDefined,
                          "int overflow: too large nHeaderOffset");
@@ -1092,9 +1093,10 @@ GDALDataset *ERSDataset::Open(GDALOpenInfo *poOpenInfo)
             for (int iBand = 0; iBand < nBands; iBand++)
             {
                 // Assume pixel interleaved.
-                auto poBand = cpl::make_unique<ERSRasterBand>(
+                auto poBand = std::make_unique<ERSRasterBand>(
                     poDS.get(), iBand + 1, poDS->fpImage,
-                    nHeaderOffset + iWordSize * iBand * poDS->nRasterXSize,
+                    nHeaderOffset + static_cast<vsi_l_offset>(iWordSize) *
+                                        iBand * poDS->nRasterXSize,
                     iWordSize, iWordSize * nBands * poDS->nRasterXSize, eType,
                     bNative);
                 if (!poBand->IsValid())
